% Gibbs Algorithm - Gibbs sampling algorithm described in 
%   Brynildsen MP, Tran LM, JC Liao "A Gibbs Sampler for the Identification of Gene Expression and Network Connectivity Consistentcy." Bioinformatics, 2006 Dec 15;22(24):3040-6. Epub 2006 Oct 23
% 
% function [consEvolve,Gene,TF] = gibbsSampling(conn,data,gc,Gene,TF,priors,bt,ConvThresh); 
%    
%     Parameters
%     ----------
%       conn: connectivity matrix (genes x tfs)
%       data: gene expression matrix (genes x experiments)
%       gc: the max number of iterations if gibbs algorithm has not converged
%       Gene: column vector of gene names corresponding to those in conn & data (genes x 1)
%       TF: column vector of TF names corresponding to those in conn (tfs x 1)
%       priors: value of 1 indicates  informative prior (main text), value of 0 indicates non-informative prior (.5)
%       bt: probability threshold for Bayesian statistic (default/main text = .9)
%       ConvThresh: convergence threshold (default/main text = .9)
%
%     Output
%     ----------
%       consEvolve: consistency log for those iterations with at least 80% the maximum number of consistent genes (MAX)
%       Gene: gene identities for those represented in consEvolve
%       TF: transcription factors ChIP-chip bound to the promoters of genes in consEvolve
%       Note: Proportional likelihood for consistency of genes can be obtained by the following command line: likelihood = sum(consEvolve');
%       Note: genes deemed consistent were those with the top MAX likelihoods
%       Note: this algorithm has not been optimized for computation speed. We suggest a number of strategies to keep computation time down: 
%           1) Analyze genes with significant expression change (2-fold change up/down in main text)
%           2) Analyze expression sets that were obtained in similar environments
%           3) To obtain rough estimates, you may decrease ConvThresh
%    
% ------------------------------------------------------------------------------------------
% Last revision December 7th, 2006 via MP Brynildsen, UCLA
% ------------------------------------------------------------------------------------------

function [consEvolve,Gene,TF] = gibbsSampling(conn,data,gc,Gene,TF,priors,bt,ConvThresh);

%Z-score data
[data] = zscore_real(data);

%Identify # Genes (N), TFs (L), Experiments (M)
[N,L] = size(conn);
[n,M] = size(data);
if (N ~= n)
    fprintf('The number of genes in your connectivity and expression data does not match, Step 1');
    stop
end

%Initialize random number generator states by clock 
rand('state',sum(100*clock))
randn('state',sum(100*clock))

%Delete genes without regulation
conn(find(conn ~= 0)) = 1;
row_sum = sum(conn');
x = find(row_sum == 0);
conn(x,:) = [];
data(x,:) = [];
Gene(x) = [];
%Delete TFs that do not regulate
x = find(sum(conn) == 0);
conn(:,x) = [];
TF(x) = [];

%Specify maximum degree of resolution
evaluation_criteria = min([M L]);

good_dist = 0;
while (good_dist == 0),
    %Populate distributions for Bayesian statistic
    [dist_full,dist_def] = bayesstat_real(conn,data);
    fprintf('Statistics Generator Complete\n');
    fprintf(1,'Bayesian threshold: %.2f\n',bt);
    
    %Calculate thresholds based on Bayesian Statistic
    [thresh_speed] = thresh(dist_full,dist_def,conn,data,bt,priors);
    tspeed = max(size(thresh_speed));
    thresh_speed
    evaluation_criteria = tspeed + 1;
    good_dist = 1;
    for jj = 1:tspeed-1,
        if (thresh_speed(jj+1) < thresh_speed(jj)),
            fprintf('Error in distributions\n');
            good_dist = 0;
        end
    end
end

%Prune Genes and TFs that cannot be resolved
fprintf('Old number of Genes: %.0f, and TFs: %.0f\n',N,L);
conn(find(conn ~= 0)) = 1;
row_sum = sum(conn');
x = find(row_sum >= evaluation_criteria-1);
conn(x,:) = [];
data(x,:) = [];
Gene(x) = [];
fprintf('Number of genes processed out due to evaluation_criteria: %.0f\n',max(size(x)));
x = find(sum(conn) == 0);
conn(:,x) = [];
TF(x) = [];
fprintf('Number of regulators now without regulation processed out: %.0f\n',max(size(x)));
[N,L] = size(conn);
[n,M] = size(data);
if (N ~= n)
    fprintf('The number of genes in your pruned connectivity and expression data does not match, Step 2');
    stop
end
fprintf('New number of Genes: %.0f, and TFs: %.0f\n',N,L);

%Randomize the nonzero values in the connectivity matrix
conn = randn(size(conn)).*conn;

if (rank(conn) < L),
    fprintf('There is a rank problem!!!\n');
    %Process data to make rank condition
    [conn,Gene,TF,data] = rank_maker(conn,Gene,TF,data);
end
[N,L] = size(conn);
[n,M] = size(data);
fprintf('New number of Genes: %.0f, and TFs: %.0f\n',N,L);
if (N ~= n)
    fprintf('The number of genes in your connectivity and expression data does not match, Step 3');
    stop
end


%Randomize the nonzero values in the connectivity matrix
conn(find(conn ~= 0)) = 1;
conn = randn(size(conn)).*conn;
x = find(abs(conn) < .1 & abs(conn) > 0);
[rx,cx] = size(x);
while (rx > 0),
    conn(x(:)) = conn(x(:))*10;
    x = find(abs(conn) < .1 & abs(conn) > 0);
    [rx,cx] = size(x);
end

%Select initial eta randomly
fprintf('Choosing Initial Eta\n');
uniq = 0;
while (uniq ~= L),
    pcount = 1;
    problem = [];
    for kk = 1:L,
        set = find(conn(:,kk) ~= 0);
        set(:,2) = randn(size(set));
        set(:,3) = set(:,1);
        set(:,1) = [];
        set = sortrows(set);
        set(:,1) = [];
        tfref(kk,1) = set(1);
        set = [];
    end
    for kk = 1:L,
        [x,xx] = size(find(tfref == tfref(kk,1)));
        if (x ~= xx),
            problem(pcount,1) = kk;
            pcount = pcount + 1;
        end
    end
    [p_row,p_col] = size(problem);
    for kk = 1:p_row,
        set = find(conn(:,problem(kk)) ~= 0);
        set(:,2) = randn(size(set));
        set(:,3) = set(:,1);
        set(:,1) = [];
        set = sortrows(set);
        set(:,1) = [];
        tfref(problem(kk),1) = set(1);
        set = [];
    end 
    uniq = size(unique(tfref,'rows'));
end

%Begin Gibbs Loop
counter = 0;
done_yet = 0;
TfReF = zeros(size(tfref));
consEvolve = [];
consFUNC = [];
the_counter = 1;
Cons_log = zeros(N,1);
while (counter < gc && done_yet == 0),
    counter = counter + 1;
    fprintf('Gibbs Counter:%.0f\n',counter);
    noise = zscore_real(randn(1,M));
    for jj = 1:L,
        previous = tfref(jj);
        tfreff = tfref;
        tfreff(jj) = [];
        connn = conn;
        connn(tfreff(:),:) = 0;
        set = find(connn(:,jj) ~= 0);
        [s,ss] = size(set);
        pivot = tfref;
        for jk = 1:s, %iterates through possible reps for TFi
            sum_real = 0;
            sum_background = 0;
            row_ref = [1:N]';
            pivot(jj,1) = set(jk);
            row_ref(pivot(:)) = [];
            beta = conn(row_ref(:),:)*inv(conn(pivot(:),:));
            beta(find(abs(beta) < 1E-10)) = 0;
            X = find(beta(:,jj) == 0);
            beta(X(:),:) = [];
            row_ref(X(:)) = [];
            [alpha_rows, alpha_columns] = size(row_ref);
            for alpha_i = 1:alpha_rows,
                keep = find(beta(alpha_i,:) ~= 0)';
                [row_e,col_e] = size(keep);
                if (row_e < evaluation_criteria-1),
                    %stat test for consistency
                    real_condition = cond(data([pivot(keep(:));row_ref(alpha_i)],:));
                    if (real_condition > thresh_speed(row_e+1)),
                        noise_condition = cond([data(pivot(keep(:)),:);noise]);
                        if (noise_condition < .9*real_condition & noise_condition < thresh_speed(row_e+1)),
                            keep(find(keep == jj)) = [];
                            noise_condition = cond([data([pivot(keep);row_ref(alpha_i)],:);noise]);   
                            if (noise_condition < .9*real_condition & noise_condition < thresh_speed(row_e+1)),
                                sum_real = sum_real + 1;
                            end
                        end
                    end
                end
            end
            characteristic(jk,1) = sum_real^2;
        end
        character_check = characteristic;
        if (sum(characteristic) == 0),
            characteristic = ones(size(characteristic));
        end
        weighting = characteristic/sum(characteristic);
        variable = rand(1);
        value1 = 0;
        for pj = 1:s, %for probability based selection of next round
            value2 = value1;
            value1 = value1 + weighting(pj);
            if (variable < value1 && variable > value2),
                tfref(jj) = set(pj);
                if (character_check(pj) > 0),
                    TfReF(jj) = 1;
                else
                    TfReF(jj) = 0;
                end
                break
            end
        end
        character_check = [];
        sum_back = 0;
        sum_n_back = 0;
        [background] = zscore_real(randn(1,M));
        row_ref = [1:N]';
        row_ref(tfref(:)) = [];
        beta = conn(row_ref(:),:)*inv(conn(tfref(:),:));
        beta(find(abs(beta) < 1E-10)) = 0;
        [alpha_rows, alpha_columns] = size(row_ref);
        cons = zeros(alpha_rows,1);
        cons_back = cons;
        for alpha_i = 1:alpha_rows,
            keep = find(beta(alpha_i,:) ~= 0)';
            [row_e,col_e] = size(keep);
            if (row_e < evaluation_criteria-1),
                %stat test for consistency
                real_cond = cond(data([tfref(keep(:));row_ref(alpha_i)],:));
                noise_cond = cond([data([tfref(keep(:))],:);background]);
                if (real_cond > thresh_speed(row_e+1) & noise_cond < .9*real_cond & noise_cond < thresh_speed(row_e+1)),
                    cons(alpha_i,1) = 1;
                end
            end
        end
        Consistent(:,jj) = zeros(N,1);
        Consistent(row_ref,jj) = cons;
        Consistent(tfref,jj) = TfReF;
        FUNC(1,jj) = sum(Consistent(:,jj));
        characteristic = [];
        for_plot(the_counter,1) = sum(Consistent(:,jj));
        the_counter = the_counter + 1;
    end
    Cons_log = Cons_log + sum(Consistent')';
    FUNC = [FUNC consFUNC];
    Consistent = [Consistent consEvolve];
    maxFUNC = max(FUNC);
    setFUNC = find(FUNC > .8*maxFUNC);
    consFUNC = FUNC(setFUNC);
    consEvolve = Consistent(:,setFUNC);
    [secn,instances] = size(setFUNC);
    fprintf('Maximum Consistency: %.0f, Number of Instances: %.0f\n',max(FUNC),instances);
    Consistent = [];
    FUNC = [];
    REG = [];
    if (instances > 1000), 
        %Log resuts to check convergence
        [logs] = convergencecheck(consEvolve,instances,.25)
        converged = find(logs < ConvThresh);
        [conv,erged] = size(converged);
        if (conv == 0),
            done_yet = 1
        end
    end
end
Gene = char(Gene);
TF = char(TF);









%-------Bayes Stat Generator------
function [dist_full,dist_def] = bayesstat_real(A,E);
[N,L] = size(A);
[n,M] = size(E);
if (n ~= N),
    fprintf('Your matrices do not match up!\n');
    stop
end

evaluation_criteria = min([L M 11]);

for kk = 2:evaluation_criteria-1,
    %Full rank
    fprintf(1,'Evaluating:\t%.0f\n',kk);
    counter = 1;       
    %Random sampling per data point
    eval = [];
    while (counter ~= 5001),
        for jj = 1:M,
            vector = rand(kk,1);
            vector = round(vector*(N-1)) + 1;
            zvector = E(vector(:),jj);
            eval(:,jj) = zvector;
        end
        [zeval] = zscore_real(eval);
        dist_full(counter,kk) = cond(zeval);
        counter = counter + 1;
    end
    %Rank Deficient
    counter = 1;
    a = A;
    e = E;
    a(find(a ~= 0)) = 1;
    sum_a = sum(a');
    del = find(sum_a >= kk);
    a(del,:) = [];
    e(del,:) = [];
    [NN,LL] = size(a);
    while (counter ~= 5001),
        choose = [];
        choose = [randn(NN,1) [1:NN]'];
        choose = sortrows(choose);
        choose(:,1) = [];
        chosen = choose(1:kk);
        r = sprank(a(chosen(:),:));
        if (r < kk),
            dist_def(counter,kk) = cond(e(chosen(:),:));
            counter = counter + 1;
        end
        chosen = [];
    end
    a = [];
    e = [];
end

%-------Z-score------
function [zdata] = zscore_real(E)
[N,M] = size(E);
for kk = 1:N,
    Mean = mean(E(kk,:));
    Std = std(E(kk,:));
    for jj = 1:M,
        zdata(kk,jj) = (E(kk,jj) - Mean)/Std;
    end
end


%-------Convergence Check--------
function [logs] = convergencecheck(consistent,instances,percentage);
[N,in] = size(consistent);
first_half = consistent(:,1:round(instances/2));
second_half = consistent(:,round(instances/2)+1:instances);
first_half = sortrows([sum(first_half')' [1:N]']);
second_half = sortrows([sum(second_half')' [1:N]']);
max_fh = max(first_half(:,1));
max_sh = max(second_half(:,1));
for kk = 1:10,
    if ((10-kk)/10 > percentage),
        first_reps = first_half(find(first_half(:,1) >= (10-kk)/10*max_fh),2); %top x percentage
        second_reps = second_half(find(second_half(:,1) >= (10-kk)/10*max_sh),2);
        [fi,re] = size(first_reps);
        [se,re] = size(second_reps);
        set = [first_reps;second_reps];
        uni = unique(set);
        [ss,sss] = size(set);
        [uu,uuu] = size(uni);
        logs(kk,1) = (ss-uu)/(min([fi se]));
    else
        first_reps = first_half(find(first_half(:,1) >= percentage*max_fh),2); %top x percentage
        second_reps = second_half(find(second_half(:,1) >= percentage*max_sh),2);
        [fi,re] = size(first_reps);
        [se,re] = size(second_reps);
        set = [first_reps;second_reps];
        uni = unique(set);
        [ss,sss] = size(set);
        [uu,uuu] = size(uni);
        logs(kk,1) = (ss-uu)/(min([fi se]));
        break
    end
end

%------------Threshold Generator-------------
function [thresh_speed] = thresh(dist_full,dist_def,conn,data,bt,priors)

[N,L] = size(conn);
[N,M] = size(data);
evaluation_criteria = min([M L 11]); 
thresh_speed = 0;

for thresh_sp = 2:evaluation_criteria-1,
    good = 0;
    [row_t,col_t] = size(thresh_speed);
    if (col_t < thresh_sp - 1),
        break
    end
    check_start = 1.2;
    base_interval = .2;
    check_interval = .2;
    while(good == 0),
        [full_value,fp] = size(find(dist_full(:,thresh_sp) > check_start-check_interval & dist_full(:,thresh_sp) < check_start+check_interval));
        [def_value,fp] = size(find(dist_def(:,thresh_sp) > check_start-check_interval & dist_def(:,thresh_sp) < check_start+check_interval));
        if (def_value > 0 && full_value > 0),
            if (priors == 1),
                prior = .5^(thresh_sp - 1);
            else
                prior = .5;
            end
            prob = def_value*prior/(def_value*prior + full_value*(1-prior));
            if (prob > bt),
                thresh_speed(1,thresh_sp) = check_start
                good = 1;
            else
                check_start = check_start + base_interval;
            end
        else
            check_start = check_start + base_interval;
            if (check_start-check_interval > max(dist_full(:,thresh_sp)) & check_start-check_interval < max(dist_def(:,thresh_sp))),
                needed = (bt*(1-prior))/(prior*(1-bt));
                max_full = max(dist_full(:,thresh_sp));
                def_found = find(dist_def(:,thresh_sp) >= max_full);
                [ppp,ooo] = size(def_found);
                if (ppp < needed),
                    good = 1;
                else
                    def_found = sort(dist_def(def_found,thresh_sp));
                    to_stop = round(needed/2);
                    check_start = def_found(to_stop);
                    thresh_speed(1,thresh_sp) = check_start;
                    good = 1;
                end
            end
            if (check_start-check_interval > max(dist_full(:,thresh_sp)) & check_start-check_interval > max(dist_def(:,thresh_sp))),
                good = 1
            end
        end
    end
end

function [A,Gene,TF,data] = rank_maker(A,Gene,TF,data)
A = randn(size(A)).*A;
Z = rank(A);
[M] = size(A);
[a,b] = size(A);
Min = min(M);
Diff = Min-Z;
if (Z == Min),
    fprintf('Linearly Independent\n\n');
end
while(Z ~= Min),
    fprintf('You have a rank problem\n\n');
    [a,b] = size(A);
    rank(A);
    for i = 1:b,
        A1 = A;
        A1(:,i) = [];
        Z1(i) = rank(A1);
    end
    [Delete] = find(Z1 == Z);
    fprintf('You must remove %.0f of the following columns:\n', Diff);
    [dd,ddd] = size(Delete);
    for jk = 1:ddd
        fprintf(1,'%s:\t%.0f\n',char(TF(Delete(jk))),Delete(jk));
    end
    for jj = 1:min(M),
        connections(jj,:) = size(find(A(:,jj) ~= 0));
    end
    fprintf('The connections for each is:\n');
    fprintf('%.0f\n', connections(Delete,1));
    Delete = sortrows([randn(ddd,1) Delete']);
    Delete(:,1) = [];
    fprintf('Removing: %.0f\n',Delete(1:Diff));
    user_entry1 = Delete(1:Diff);
    [x,y] = find(A(:,user_entry1(:)) ~= 0);
    [r,s] = size(x);
    A(:,user_entry1(:)) = [];
    TF(user_entry1) = [];
    if (r > 0),
        A(x(:),:) = [];
        Gene(x) = [];
        data(x,:) = [];
    end
    Z = rank(A);
    [M] = size(A);
    Min = min(M);
    Diff = Min-Z;
    Z1 = [];
end